#include "main.h"
#include "File.h"

namespace FileTMP {
    void help() {
	printf(".tmp files. These are an ImageStack format used for intermediate results.\n"
	       "They support any number of of frames and channels, and are stored as a raw\n"
	       "32 bit float arrays with a 4 int (16 byte) binary header specifying the\n"
	       "dimensions.\n\n");
    }

    void save(Window im, string filename) {
        FILE *f = fopen(filename.c_str(), "wb");
        assert(f, "Could not write output file %s\n", filename.c_str());
        // write the dimensions
        fwrite(&im.frames, sizeof(int), 1, f);
        fwrite(&im.width, sizeof(int), 1, f);
        fwrite(&im.height, sizeof(int), 1, f);
        fwrite(&im.channels, sizeof(int), 1, f);
        // write the data
	for (int t = 0; t < im.frames; t++) {
	    for (int y = 0; y < im.height; y++) {
		fwrite(im(t, 0, y), sizeof(float), im.width * im.channels, f);
	    }
	}
        fclose(f);
    }

    Image load(string filename) {
        FILE *file = fopen(filename.c_str(), "rb");
        assert(file, "Could not open file %s\n", filename.c_str());
        
        // get the dimensions
	int frames, width, height, channels;
        fread(&frames, sizeof(int), 1, file);
        fread(&width, sizeof(int), 1, file);
        fread(&height, sizeof(int), 1, file);
        fread(&channels, sizeof(int), 1, file);
	Image im(frames, width, height, channels);

        // get the data
        fread(im(0, 0, 0), sizeof(float), width * height * frames * channels, file);
        fclose(file);

	return im;
    }
}
